<?php

namespace Drupal\menu_migration\Service;

use Drupal\Component\Serialization\Json;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Config\ImmutableConfig;
use Drupal\Core\DrupalKernel;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Menu\MenuLinkTreeInterface;
use Drupal\Core\Menu\MenuTreeParameters;
use Drupal\menu_link_content\Entity\MenuLinkContent;
use Drupal\menu_link_content\Plugin\Menu\MenuLinkContent as MenuLinkContentPlugin;
use Drupal\menu_migration\Event\MenuImportEvent;
use Drupal\menu_migration\Form\SettingsForm;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Contracts\EventDispatcher\EventDispatcherInterface;

/**
 * Service for managing menu items migrations.
 */
class MenuMigrationService {

  /**
   * The entity type manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * The menu migration settings config.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected ImmutableConfig $config;

  /**
   * The file system service.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected FileSystemInterface $fileSystem;

  /**
   * The menu link tree service.
   *
   * @var \Drupal\Core\Menu\MenuLinkTreeInterface
   */
  protected MenuLinkTreeInterface $menuTree;

  /**
   * The current request.
   *
   * @var \Symfony\Component\HttpFoundation\Request
   */
  protected $currentRequest;

  /**
   * The event dispatcher.
   *
   * @var \Symfony\Contracts\EventDispatcher\EventDispatcherInterface
   */
  protected EventDispatcherInterface $eventDispatcher;

  /**
   * Constructs a new menu items migration service.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $configFactory
   *   The config factory.
   * @param \Drupal\Core\File\FileSystemInterface $fileSystem
   *   The file system service.
   * @param \Drupal\Core\Menu\MenuLinkTreeInterface $menuLinkTree
   *   The menu tree service.
   * @param \Symfony\Component\HttpFoundation\RequestStack $request
   *   The request stack service.
   * @param \Symfony\Contracts\EventDispatcher\EventDispatcherInterface $eventDispatcher
   *   The event dispatcher service.
   */
  public function __construct(EntityTypeManagerInterface $entityTypeManager, ConfigFactoryInterface $configFactory, FileSystemInterface $fileSystem, MenuLinkTreeInterface $menuLinkTree, RequestStack $request, EventDispatcherInterface $eventDispatcher) {
    $this->entityTypeManager = $entityTypeManager;
    $this->config = $configFactory->get(SettingsForm::SETTINGS);
    $this->fileSystem = $fileSystem;
    $this->menuTree = $menuLinkTree;
    $this->currentRequest = $request->getCurrentRequest();
    $this->eventDispatcher = $eventDispatcher;
  }

  /**
   * Gets the available menus.
   *
   * @return array
   *   Returns the available menus as associative array, where the key is the ID
   *   and the value is the label.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function getAvailableMenus() {
    $menus = [];
    $availableMenus = $this->entityTypeManager->getStorage('menu')->loadMultiple();
    foreach ($availableMenus as $menu) {
      $menus[$menu->id()] = $menu->label();
    }
    return $menus;
  }

  /**
   * Exports the menu link content items from the given menu to json.
   *
   * @param string $menuName
   *   The menu machine name.
   *
   * @return bool
   *   Returns TRUE if successful, FALSE otherwise.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function exportMenu(string $menuName) {
    $directory = $this->getTargetDirectory($menuName);
    if (!$this->fileSystem->prepareDirectory($directory, FileSystemInterface::CREATE_DIRECTORY | FileSystemInterface::MODIFY_PERMISSIONS)) {
      return FALSE;
    }

    $bytes = file_put_contents("{$directory}/{$menuName}.json", Json::encode($this->getMenuTree($menuName)));
    return !empty($bytes);
  }

  /**
   * Imports the menu from the exported location.
   *
   * @param string $menuName
   *   The menu name.
   *
   * @return bool
   *   Returns TRUE if successful, FALSE otherwise.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function importMenu(string $menuName) {
    // Delete existing items.
    $existingItems = $this->entityTypeManager->getStorage('menu_link_content')->loadByProperties(['menu_name' => $menuName]);
    foreach ($existingItems as $existingItem) {
      $existingItem->delete();
    }

    $directory = $this->getTargetDirectory($menuName);
    // @todo Check if we can read.
    if ($contents = file_get_contents("{$directory}/{$menuName}.json")) {
      $items = Json::decode($contents);
      $this->generateMenuItems($items);
      return TRUE;
    }
    return FALSE;
  }

  /**
   * Gets the target directory for the given menu name.
   *
   * @param string $menuName
   *   The menu name.
   *
   * @return string
   *   Returns the full path to the target directory.
   */
  public function getTargetDirectory($menuName) {
    $path = [
      DRUPAL_ROOT,
      $this->config->get('export_path'),
      DrupalKernel::findSitePath($this->currentRequest),
      $menuName,
    ];
    return implode('/', $path);
  }

  /**
   * Generates menu items and stores them in the database.
   *
   * @param array $sourceItems
   *   The hierarchy of menu items with Menu link content properties.
   * @param int|string $parent
   *   The menu link parent. Defaults to 0.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  protected function generateMenuItems(array $sourceItems, $parent = 0) {
    foreach ($sourceItems as $item) {
      // Allow others to alter an item right before save. Does not include
      // translations.
      $menuItemEvent = new MenuImportEvent($item);
      $this->eventDispatcher->dispatch($menuItemEvent, MenuImportEvent::IMPORT_MENU_ITEM);
      $item = $menuItemEvent->getMenuItem();
      $item['parent'] = $parent;
      $translations = $item['translations'] ?? [];
      unset($item['translations']);
      $menuLinkContent = MenuLinkContent::create($item);
      $menuLinkContent->save();
      foreach ($translations as $langcode => $translation) {
        $menuLinkContent->addTranslation($langcode, $translation)->save();
      }
      if (!empty($item['children'])) {
        $this->generateMenuItems($item['children'], "menu_link_content:{$menuLinkContent->uuid()}");
      }
    }
  }

  /**
   * Gets the menu tree for the given menu name.
   *
   * @param string $menuName
   *   The menu name.
   *
   * @return array
   *   Returns an array with the menu hierarchical structure.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  protected function getMenuTree($menuName) {
    $parameters = new MenuTreeParameters();
    $tree = $this->menuTree->load($menuName, $parameters);
    $manipulators = [
      ['callable' => 'menu.default_tree_manipulators:generateIndexAndSort'],
    ];
    $tree = $this->menuTree->transform($tree, $manipulators);
    $processedTree = [];
    $this->processTree($tree, $processedTree);
    return $processedTree;
  }

  /**
   * Processes the given menu tree.
   *
   * Converts the given menu tree in a tree of hierarchical items, where each
   * item is a set of menu link properties.
   *
   * @param \Drupal\Core\Menu\MenuLinkTreeElement[] $tree
   *   The generated menu tree.
   * @param array $processedTree
   *   The resulted array.
   * @param int $delta
   *   The delta used for building the hierarchy.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  protected function processTree(array &$tree, array &$processedTree, $delta = 0) {
    foreach ($tree as &$element) {
      $definition = $element->link->getPluginDefinition();
      if ($element->link instanceof MenuLinkContentPlugin) {
        // Unfortunately, we don't have access to the menu link content which
        // exists in the menu link content plugin, but it's protected.
        /** @var \Drupal\menu_link_content\Entity\MenuLinkContent $menuLinkContent */
        $menuLinkContent = $this->entityTypeManager->getStorage('menu_link_content')->load($definition['metadata']['entity_id']);
        $translations = [];
        $translationLanguages = $menuLinkContent->getTranslationLanguages(FALSE);
        foreach ($translationLanguages as $translationLanguage) {
          $langcode = $translationLanguage->getId();
          if ($menuLinkContent->hasTranslation($langcode)) {
            $menuItemTranslation = $menuLinkContent->getTranslation($langcode);
            foreach ($this->getTranslatableFields($menuLinkContent) as $translatableField) {
              $translations[$langcode][$translatableField] = $menuItemTranslation->get($translatableField)->getValue();
            }
          }
        }
        $processedTree[$delta] = [
          'url' => $definition['url'],
          // @todo We kinda need to load the tree in the default language.
          // Workaround: the plugin definition comes in non-default language,
          // need to investigate that.
          'title' => $menuLinkContent->get('title')->getString(),
          'description' => $menuLinkContent->get('description')->getString(),
          'options' => $definition['options'],
          'enabled' => $definition['enabled'],
          'expanded' => $definition['expanded'],
          'weight' => $definition['weight'],
          'menu_name' => $definition['menu_name'],
          'link' => $menuLinkContent->get('link')->getValue()[0],
          'translations' => $translations,
        ];
        if ($element->hasChildren) {
          $processedTree[$delta]['children'] = [];
          $this->processTree($element->subtree, $processedTree[$delta]['children']);
        }
        $delta++;
      }
    }
  }

  /**
   * Gets the menu link content translatable field names.
   *
   * @param \Drupal\menu_link_content\Entity\MenuLinkContent $menuLinkContent
   *   A menu link content entity.
   *
   * @return int[]|string[]
   *   A list of translatable fields with some filtered out.
   */
  protected function getTranslatableFields(MenuLinkContent $menuLinkContent) {
    // @todo For now we're skipping these fields that are marked as translatable. Will deal with them later.
    $skippedFields = [
      'langcode',
      'changed',
      'default_langcode',
      'revision_translation_affected',
      'metatag',
      'content_translation_source',
      'content_translation_outdated',
      'content_translation_uid',
      'content_translation_status',
      'content_translation_created',
    ];
    return array_diff(array_keys($menuLinkContent->getTranslatableFields()), $skippedFields);
  }

}
